from itertools import chain
from pymorphy2 import MorphAnalyzer
morph = MorphAnalyzer()


def inflect(msg):
    if '"' in msg:
        subj, obj, *body = msg.rstrip('.').split()
        gender = morph.parse(subj)[0].tag.gender
        for i, word in enumerate(body[:]):
            parse = morph.parse(word)[0]
            body[i] = word if not parse.tag.POS in ('VERB', 'PRTS') \
                else parse.inflect({gender}).word
        return ' '.join(chain([subj, obj], body))



from django.conf import settings
from django.core.mail import send_mail
from django.utils.text import capfirst


class Communicator:
    display_html = True

    def __init__(self, user):
        self.user = user

    def communicate(self, text):
        raise NotImplementedError

    def send(self, request, subject, component, **props):
        self.request = request
        return self.communicate(
            subject,
            component().as_string(request, **dict(
                {'user': self.user, 'subject': subject}, **props)))


class EmailCommunicator(Communicator):
    def get_from(self):
        return '{0} <{1}>'.format(str(settings.TITLE), settings.EMAIL_HOST_USER)

    def get_connection(self):
        return {}

    def communicate(self, subj, text):
        return send_mail(
            capfirst(subj),
            '' if not self.display_html else text,
            self.get_from(),
            [self.user.email],
            **dict(
                {} if not self.display_html else {'html_message': text},
                **self.get_connection()))


class CommunicationMixin:
    @property
    def comm(self):
        if not hasattr(self, '_communicator'):
            self._communicator = self.__class__.get_communicator(user=self)
        return self._communicator

    @classmethod
    def get_communicator(cls, **kw):
        if cls.USERNAME_FIELD == 'email':
            return EmailCommunicator(**kw)



from functools import reduce
from itertools import chain
from urllib.parse import unquote

from django.contrib.contenttypes.forms import BaseGenericInlineFormSet
from django.core.urlresolvers import reverse
from django.core.exceptions import ObjectDoesNotExist, FieldDoesNotExist
from django.db.models import OneToOneField
from django.forms import Form as EmptyForm
from django.forms.fields import DateField, DateTimeField, TimeField
from django.forms.models import modelform_factory, modelformset_factory, inlineformset_factory
from django.http import Http404
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.generic.edit import FormMixin

from ..forms import FrozenSelect, DateTimePicker
from ..components import Form
from ..utils.functools import just, call, swap, throw, \
    pluck, first, select, separate, factual, case, omit, pick, map_keys, map_values
from ..utils.translation import _, inflect, get_string, get_model_names
from .base import BaseView


class FormFieldsOverridesMixin:
    def get_formfields_overrides(self):
        return {
            DateField: (DateTimePicker, {'key': 'date'}),
            DateTimeField: (DateTimePicker, {'key': 'datetime'}),
            TimeField: (DateTimePicker, {'key': 'time'})}

    def override_form_fields(self, form):
        overrides = self.get_formfields_overrides()
        for field in form.fields.values():
            match = overrides.get(field.__class__, None)
            if match:
                Widget, kw = match
                field.widget = Widget(**kw)


class ModelFormMixin(FormFieldsOverridesMixin):
    empty_form = False

    @cached_property
    def model(self):
        return getattr(
            self,
            'model_class',
            hasattr(self.form_class, '_meta') \
            and self.form_class._meta.model or None)

    @cached_property
    def instance(self):
        return call(getattr(self, 'get_instance', lambda: None))

    def get_model_name(self, key='single', fn=get_string, model=None):
        return case(key, dict(
            zip(('single', 'plural'),
            map(fn, get_model_names(model or self.model)))))

    def get_model_fields(self, model=None):
        return pluck((model or self.model)._meta.fields[1:], 'name')

    def get_fields(self):
        return self.form_class and self.form_class._meta.fields \
            or getattr(self, 'fields', self.get_model_fields())

    def get_choices(self, model, field_name):
        field = model._meta.get_field(field_name)
        if getattr(field, 'rel', None):
            return [(obj.pk, str(obj)) for obj in field.rel.to.objects.all()]
        return list(field.choices)

    def get_relation_fields(self):
        result = []
        if self.model:
            for field_name in self.get_fields():
                try:
                    field = self.model._meta.get_field(field_name)
                    if isinstance(field, OneToOneField):
                        result.append(field)
                except FieldDoesNotExist:
                    pass
        return result

    def get_form_class(self):
        if self.form_class:
            return self.form_class
        elif self.empty_form:
            return EmptyForm
        else:
            return modelform_factory(self.model, fields=self.get_fields())

    def get_form_data(self, **kw):
        result = pick(kw, 'prefix', 'initial', 'instance')
        if self.request.POST:
            result['data'] = self.request.POST
        if self.request.FILES:
            result['files'] = self.request.FILES
        if self.empty_form:
            return omit(result, 'instance')
        return result

    def get_form(self):
        form = self.get_form_class()(**self.get_form_kwargs())
        for field in pluck(self.get_relation_fields(), 'name'):
            form.fields.pop(field)
        return self.prepare_form(form)

    def get_relation(self, field, **kw):
        kw = kw or {'fields': self.get_model_fields(field.rel.to)}
        form = modelform_factory(field.rel.to, **kw)(
            **self.get_form_data(
                prefix=field.name,
                instance=getattr(self.instance, field.name, None)))
        return {'field': field, 'form': form, 'required': not field.blank}

    def get_relations(self):
        return factual(map(
            lambda field: swap(field, getattr(
                self,
                'get_{0}_relation'.format(field.name),
                self.get_relation)),
            self.get_relation_fields()))

    def prepare_form(self, form):
        if not self.empty_form:
            self.override_form_fields(form)
        return form


class FormsetsMixin(ModelFormMixin):
    def get_formsets(self):
        return []

    def get_formset_factory(self, formset_model, **kw):
        if 'matrix' in kw:
            max_num = len(self.get_choices(formset_model, kw['matrix']))
            kw = dict({k: max_num for k in kw.pop('auto_num', ('extra', 'max_num'))}, **kw)
            kw['widgets'] = {kw.pop('matrix'): FrozenSelect()}
        if not 'fields' in kw and not 'form' in kw:
            kw['fields'] = self.get_model_fields(formset_model)
        if not 'constructor' in kw:
            kw['parent_model'] = kw.pop('model', self.model)
        if hasattr(formset_model, 'content_object'):
            ct_field = kw.pop('ct_field', 'content_type')
            fk_field = kw.pop('fk_field', 'object_id')
            result = modelformset_factory(
                formset_model,
                formset=BaseGenericInlineFormSet,
                exclude=(ct_field, fk_field),
                **omit(kw, 'parent_model', 'for_concrete_model'))
            options = formset_model._meta
            result.ct_field = options.get_field(ct_field)
            result.ct_fk_field = options.get_field(fk_field)
            result.for_concrete_model = kw.get('for_concrete_model', True)
            return result
        return kw.pop('constructor', inlineformset_factory)(model=formset_model, **kw)

    def get_formset_data(self, factory, **kw):
        result = self.get_form_data(instance=self.instance, **kw)
        if 'matrix' in kw:
            exclude = [] if not hasattr(factory, 'fk') else list(map(
                lambda obj: getattr(obj, kw['matrix']),
                factory.model.objects.filter(**{factory.fk.name: self.instance})))
            result['initial'] = list(map(
                lambda choice: {kw['matrix']: choice},
                filter(
                    lambda choice: not choice in exclude,
                    map(first, self.get_choices(factory.model, kw['matrix'])))))
        return result

    def inline_formset(self, formset_model, **kw):
        factory = self.get_formset_factory(formset_model, **kw.copy())
        prefix = self.get_model_name('plural', model=factory.model)
        get_data = getattr(self, 'get_{0}_formset_data'.format(prefix), self.get_formset_data)
        formset = factory(**get_data(factory, **dict(omit(kw, 'constructor'), prefix=prefix)))

        for form in formset.forms:
            self.override_form_fields(form)
        if formset.can_delete and formset.validate_min:
            for form in formset.forms[:formset.min_num]:
                form.fields['DELETE'].widget.attrs['disabled'] = True
        return formset


class FormView(BaseView, FormsetsMixin, FormMixin):
    use_prefix = False
    long_term = False
    components = {'content': Form}
    form_props = {}
    success_redirect = 'index'
    success_message = None
    actions = [{
        'icon': 'check-square',
        'title': _('Save'),
        'level': 'success'}]

    def get_attrs_dict(self, *attrs):
        return dict(map(lambda x: (x, call(getattr(self, 'get_' + x))), attrs))

    def get_prefix(self):
        return self.prefix or (self.use_prefix and self.about or None)

    def get_form_kwargs(self):
        return self.get_form_data(**dict(
            {} if not self.instance else {'instance': self.instance},
            **self.get_attrs_dict('prefix', 'initial')))

    def get_members(self):
        result = self.get_attrs_dict('form', 'relations', 'formsets')
        return result

    def get_all_forms(self, **kw):
        return chain(
            [kw['form']],
            pluck(kw['relations'], 'form'),
            *pluck(kw['formsets'], 'forms'))

    def get_media(self, **kw):
        return reduce(
            lambda x, y: x and (x + y) or y,
            pluck(self.get_all_forms(**kw), 'media'))

    def get_actions(self):
        return self.actions

    def get_form_props(self):
        result = self.form_props
        result['actions'] = self.get_actions()
        if self.check_is_long():
            result['spinner'] = 'spinner'
        return result

    def get_context_data(self, **kw):
        members = kw or self.get_members()
        return dict(
            super().get_context_data(**members),
            **map_keys(
                lambda key: key if not self.use_prefix else '_'.join(
                    [self.get_prefix(), key]),
                dict(
                    members,
                    props=self.get_form_props())))

    def get_success_message(self, **kw):
        return self.success_message

    def get_success_redirect(self, **kw):
        return reverse(getattr(self, 'success_redirect', self.url_match.url_name))

    def success_finish(self, **kw):
        self.send_message('success', self.get_success_message(**kw))
        return redirect(self.get_success_redirect(**kw))

    def save_relations(self, form, *relations):
        for rel in relations:
            if rel['form'].has_changed():
                setattr(form.instance, rel['field'].name, rel['form'].save())

    def save_form(self, form):
        hasattr(form, 'instance') and form.instance.save()

    def save_formsets(self, form, *formsets):
        for formset in formsets:
            formset.instance = form.instance
            formset.save()

    def process_valid(self, **kw):
        self.save_relations(kw['form'], *kw['relations'])
        self.save_form(kw['form'])
        self.save_formsets(kw['form'], *kw['formsets'])
        return self.success_finish(**kw)

    def process_invalid(self, **kw):
        errors = chain(*map(call, pluck(
            self.get_all_forms(**kw), 'non_field_errors')))
        for message in errors:
            self.send_message('error', message)
        return self.render_to_response(self.get_context_data(**kw), status=400)

    def check_is_long(self):
        return self.long_term

    def check_is_valid(self, **kw):
        return all(map(lambda x: x.is_valid(), self.get_all_forms(**kw)))

    def process(self, **kw):
        if self.check_is_valid(**kw):
            map(
                lambda x: x.save(commit=False),
                chain(
                    hasattr(kw['form'], 'save') and [kw['form']] or [],
                    *pluck(kw['relations'], 'form')))
            return self.process_valid(**kw)
        return self.process_invalid(**kw)

    def post(self, request, *args, **kw):
        return self.process(**self.get_members())


class MatrixView(FormView):
    empty_form = True

    def get_formsets(self):
        return list(map(
            lambda matrix: self.inline_formset(
                matrix[0],
                constructor=modelformset_factory,
                matrix=matrix[1]),
            self.models))

    def save_formsets(self, form, *formsets):
        for formset in formsets:
            formset.save()


class ObjectView(FormView):
    def get_instance(self):
        pk = self.url_match.kwargs.get('pk')
        try:
            return pk and self.model.objects.get(pk=pk)
        except ObjectDoesNotExist:
            raise Http404

    def get_form_props(self):
        return dict(
            super().get_form_props(),
            panels=True,
            title=self.instance and str(self.instance) or self.model._meta.verbose_name)

    def get_success_redirect(self, **kw):
        if 'next' in self.request.GET:
            return unquote(self.request.GET['next'])
        return reverse(self.get_model_name('plural'))

    def get_success_message(self, **kw):
        return inflect(_(
            'The %(name)s "%(obj)s" was changed successfully.' \
            if self.url_match.kwargs else \
            'The %(name)s "%(obj)s" was added successfully.') % {
                'name': self.get_model_name(fn=just),
                'obj': str(kw['form'].instance)})



import string
from django.conf import settings
from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import get_hasher, make_password
from django.forms import PasswordInput
from django.utils.text import capfirst
from era import _, random_str
from era.utils.functools import pick

from .components import InviteNotification, RegistrationNotification
from .decorators import anonymous_required
from .models import Confirm


class AnonymousMixin:
    decorators = [anonymous_required]
    keywords = ['auth']
    form_props = {'class': 'condensed'}


class UserMixin:
    model = get_user_model()
    extra_fields = ['password']
    validate_unique = True

    def get_fields(self):
        return [self.model.USERNAME_FIELD] + self.extra_fields

    def prepare_form(self, form):
        form = super().prepare_form(form)
        if not self.validate_unique:
            form.validate_unique = lambda: None
        return form


class PasswordMixin:
    long_term = True
    password_messages = {
        'new': _('enter new password'),
        'blank': _('leave blank for random')}

    def gen_password(self, **kw):
        return random_str(5, ''.join([
            string.ascii_uppercase,
            string.digits]), **kw)

    def prepare_form(self, form):
        form = super().prepare_form(form)
        if self.instance:
            form.fields['password'].required = False
            form.fields['password'].widget = PasswordInput({
                'placeholder': self.password_messages['new']})
        else:
            form.fields['password'].required = False
            form.fields['password'].widget = PasswordInput({
                'placeholder': self.password_messages['blank']})
        return form

    def process_valid(self, form, **kw):
        if not self.instance and not 'password' in form.changed_data:
            form.cleaned_data['password'] = self.gen_password()
        return super().process_valid(form=form, **kw)

    def save_form(self, form):
        if not self.instance or 'password' in form.changed_data:
            form.instance.set_password(form.cleaned_data['password'])
        super().save_form(form)


class LoginMixin:
    def get_access(self, user):
        return user.access

    def process_login(self, user):
        if self.get_access(user):
            auth.login(self.request, user)
            return self.success_finish()
        else:
            self.send_message('error', _('sorry, access denied'))
            return self.navigate('login')


class WelcomeMixin:
    def get_success_message(self, **kw):
        return _('welcome {username}').format(username=self.request.user.get_short_name())


class InvitationMixin(UserMixin, PasswordMixin):
    notification = InviteNotification
    notification_message = _('invitation')

    def get_user(self, form):
        return form.instance.pk and form.instance or self.model.objects \
            .filter(**pick(form.cleaned_data, self.model.USERNAME_FIELD)) \
            .first()

    def send_notification(self, form, **kw):
        if not 'password' in form.changed_data:
            kw['password'] = form.cleaned_data['password']
        self.get_user(form).comm.send(
            self.request, self.notification_message, self.notification, **kw)

    def save_form(self, form):
        new_user = not bool(form.instance.pk)
        super().save_form(form)
        if new_user:
            self.send_notification(form)


class ConfirmationMixin(InvitationMixin):
    confirm = 'password'
    back_url = 'login'
    notification = RegistrationNotification
    notification_message = _('registration')
    success_message = _('confirmation data has been sent')

    def get_actions(self):
        return super().get_actions() or [
            {'icon': 'send', 'title': _('request'), 'level': 'success'},
            {'icon': 'chevron-left', 'title': _('back'), 'level': 'link', 'link': self.back_url}]

    def gen_code(self, salt=None):
        hasher = get_hasher()
        generate = True
        while generate:
            code = random_str()
            encoded = hasher.encode(code, salt or hasher.salt())
            generate = bool(Confirm.objects.filter(code=encoded).count())
        return code, encoded

    def create_confirmation(self, user, key, password):
        code, encoded = self.gen_code(password)
        Confirm.objects.create(
            user=user,
            key=key,
            code=encoded,
            sign=make_password(password))
        return code

    def get_confirmation(self, user):
        return Confirm.objects.filter(user=user, key=self.confirm)

    def send_notification(self, form, **kw):
        kw['code'] = self.create_confirmation(
            self.get_user(form), self.confirm, form.cleaned_data['password'])
        super().send_notification(form, **kw)


class SignMixin(ConfirmationMixin):
    def get_success_redirect(self, **kw):
        result = super().get_success_redirect(**kw)
        if not self.instance and not 'password' in kw['form'].changed_data:
            return result + '?sign=blank'
        return result



from django.contrib import auth
from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from era.utils.functools import unidec, omit


@unidec
def role_required(method, req, *args, **kw):
    if req.user.role in kw.get('allow', []):
        return method(req, *args, **omit(kw, 'allow'))
    raise PermissionDenied()


@unidec
def anonymous_required(method, req, *args, **kw):
    if req.user.is_authenticated():
        if kw.get('logout', False):
            auth.logout(req)
            return redirect(req.get_full_path())
        else:
            return redirect('/')
    return method(req, *args, **omit(kw, 'logout'))



from django.conf import settings
from django.utils.text import capfirst
from era import _
from era.components import register, TemplateComponent, Tag, Component, Break, Link, Image, Menu
from era.utils.functools import factual, pick
from era.utils.translation.verbose import verbose_attr


class UserModelMixin:
    @property
    def is_superuser(self):
        return self.request.user.is_authenticated() \
            and self.request.user.role == settings.USER_ROLES[0].string

    def get_user_menu_item(self):
        return {
            'model': 'users.User',
            'caption': {'icon': 'users'},
            'include': self.is_superuser}


class AuthMenu(Menu):
    def get_items(self):
        if self.request.user.is_anonymous():
            return [{
                'caption': {'icon': 'sign-in', 'title': _('login')},
                'link': {'url': 'login'},
                'active': self.check_active('login', 'join', 'reset', 'unlock', 'confirm')}]
        return [{
            'caption': {'icon': 'user', 'title': _('profile')},
            'link': {'url': 'profile'}
            }, {
            'caption': {'icon': 'sign-out', 'title': _('logout')},
            'link': {'url': 'logout'}}]


class BaseUsersMenu(UserModelMixin, AuthMenu):
    def get_items(self):
        return [self.get_user_menu_item()] + super().get_items()


@register.era
class Logo(Link):
    inline = True

    def resolve_props(self):
        return {
            'host': True,
            'url': 'index',
            'newtab': True}

    def get_nodelist(self):
        return self.inject(Image, {
            'host': True,
            'url': 'images/logo.png'})


class Notification(Component):
    parts = []

    def get_parts(self):
        return self.parts

    def render_empty(self):
        return ' '

    def render_value(self, label, value=None):
        return ': '.join([capfirst(str(label)), value or self.props.get(label.string)])

    def render_logo(self):
        if self.props.user.comm.display_html:
            return self.inject(Logo)

    def render_link(self, **kw):
        if self.props.user.comm.display_html:
            return self.inject(Link, dict({'host': True, 'newtab': True}, **kw))

    def DOM(self):
        delimiter = Break().as_string() if self.props.user.comm.display_html else '\n'
        return delimiter.join(factual(map(
            lambda attr: getattr(self, '_'.join(['render', attr]))(),
            self.get_parts())))


class AuthNotification(Notification):
    def get_parts(self):
        result = ['logo', 'empty']
        if 'code' in self.props:
            result.append('code')
        if 'password' in self.props:
            result.append('password')
        return result + ['link']

    def render_code(self):
        return self.render_value(_('code'))

    def render_password(self):
        return self.render_value(_('password'))

    def render_link(self, **kw):
        qs = pick(self.props, 'code')
        if 'password' in self.props and 'code' in qs:
            qs['sign'] = self.props.password
        if not 'nodelist' in kw:
            kw['nodelist'] = _('confirm')
        return super().render_link(url=self.url_name, qs=qs, **kw)


class InviteNotification(AuthNotification):
    url_name = 'login'

    def render_link(self, **kw):
        return super().render_link(nodelist=_('login'), **kw)


class RegistrationNotification(AuthNotification):
    url_name = 'confirm'


class ResetNotification(AuthNotification):
    url_name = 'unlock'



from django.views.generic.base import RedirectView
from .base import BaseViewMixin, BaseView, DisplayAttrMixin
from .forms import FormView, MatrixView, ObjectView
from .lists import ListView, CollectionView



from .utils.functools import just, call, swap, unpack_args, throw, unidec, \
    pluck, factual, pick, omit, truthful, avg, random_str
from .utils.translation import _
from .utils.urls import view_url, admin_urls, package_patterns



from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, BaseUserManager
from django.core.validators import RegexValidator
from django.db import models
from era.utils.translation import _, verbose_choices
from .communicators import CommunicationMixin


class Confirm(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL)
    key = models.CharField(max_length=30)
    code = models.CharField(max_length=128, verbose_name=_('code'), unique=True)
    sign = models.CharField(max_length=128, verbose_name=_('confirmation'))


class BaseUser(CommunicationMixin, AbstractBaseUser):
    class Meta:
        abstract = True
        verbose_name = _('user')
        verbose_name_plural = _('users')

    USERNAME_FIELD = 'email'
    objects = BaseUserManager()
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('date joined'))
    access = models.BooleanField(_('access'), default=True)

    email = models.EmailField(verbose_name=_('email'), unique=True, null=True)
    name = models.CharField(verbose_name=_('username'), max_length=20, null=True)
    role = models.CharField(
        verbose_name=_('role'),
        max_length=10,
        choices=verbose_choices(*settings.USER_ROLES))

    @property
    def username_value(self):
        return getattr(self, self.USERNAME_FIELD)

    @property
    def username_dict(self):
        return {self.USERNAME_FIELD: self.username_value}

    def __str__(self):
        return '{0} <{1}>'.format(self.name, self.username_value)

    def get_short_name(self):
        return self.name



from itertools import chain
from django.conf import settings
from django.core.urlresolvers import resolve

from ..utils.functools import call, unpack_args, factual, just, pick, first
from ..utils.translation.string import _
from .library import Component, Tag, ScriptedTag
from .markup import MarkedList, Break, Link, Icon, Caption, Column, Panel, Table
from .forms import Action


class QuerySetKey(Link):
    def get_defaults(self):
        return dict(super().get_defaults(), replace=False, collapse=False, clean=[])

    def get_url(self):
        qd = self.request.GET.copy()
        if self.props.replace:
            for key in qd.copy().keys():
                if key.startswith(self.props.method):
                    qd.pop(key)
        if self.props.value is None:
            self.props.clean.append(self.props.argument)
        else:
            qd[self.props.argument] = self.props.value
        for arg in self.props.clean:
            arg in qd and qd.pop(arg)
        return '?'.join([self.request.path, qd.urlencode()]).rstrip('?')

    def resolve_props(self):
        return {
            'reverse': False,
            'argument': '-'.join(factual([self.props.method, self.props.key]))}

    def resolve_attrs(self):
        if (self.props.value is None and not self.props.argument in self.request.GET) \
        or str(self.props.value) == self.request.GET.get(self.props.argument):
            return {}
        return super().resolve_attrs()

    def DOM(self):
        if not self.props.attrs:
            if not self.props.collapse:
                return self.props.nodelist
            return ''
        return super().DOM()


class QuerySetFilter(Panel):
    def render_key(self, value, name, count=None):
        return self.inject(
            QuerySetKey, dict({
                'method': 'filter',
                'key': self.props.key,
                'clean': ['page'],
                'value': value,
                'collapse': value is None}),
            ' '.join([
                str(name),
                '({0})'.format(count) if self.props.counters and count else '']))

    def render_clear(self):
        if not self.props.get('ignore_length', False) and len(self.props.choices) == 1:
            return ''
        return self.render_key(None, self.inject(
            Icon, {'name': 'remove', 'class': 'qs-clear pull-right'}))

    def render_title(self):
        return ''.join([str(self.props.title), self.render_clear()])

    def render_keys(self):
        return self.inject(
            Break, {'join': map(
                unpack_args(self.render_key),
                self.props.choices)})

    def resolve_props(self):
        return dict(super().resolve_props(), **{
            'title': self.render_title(),
            'body': self.render_keys()})


class SortableTable(Table):
    def render_arrow(self, title, key):
        if not key:
            return title
        method = 'sort'
        value = self.request.GET.get('-'.join([method, key]))

        if value is None:
            value = True
            icon = None
        elif value == 'True':
            value = False
            icon = 'angle-double-down'
        else:
            value = True
            icon = 'angle-double-up'
        return self.inject(
            QuerySetKey,
            {'method': method, 'key': key, 'value': value, 'replace': True},
            title if icon is None else self.inject(
                Caption, {'title': title, 'icon': icon}))

    def render_content(self, content, cell):
        if cell == 'th':
            return self.render_arrow(*content)
        return super().render_content(content, cell)


class Paginator(MarkedList):
    def resolve_props(self):
        return {'page': self.context['page_obj'], 'class': 'pager'}

    def render_arrow(self, control, direction):
        if not call(getattr(self.props.page, 'has_' + control)):
            return ''
        return self.inject(
            QuerySetKey,
            {
                'method': 'page',
                'key': '',
                'value': call(getattr(self.props.page, control + '_page_number'))},
            self.inject(Icon, {'name': 'angle-double-' + direction}))

    def get_items(self):
        return [
            self.render_arrow('previous', 'left'),
            self.inject(
                Tag, {'el': 'span'}, ' / '.join(map(str, [
                    self.props.page.number,
                    self.props.page.paginator.num_pages]))),
            self.render_arrow('next', 'right')]


class SearchLine(ScriptedTag):
    el = 'input'
    nobody = True

    def resolve_attrs(self):
        return {'class': 'form-control', 'placeholder': _('Search')}


class ChangeList(Component):
    def resolve_props(self):
        return dict(self.context.dicts[2], **settings.ERA_COLLECTIONS)

    def get_location_qs(self):
        return {'next': self.request.get_full_path()}

    def render_objects(self):
        return list(map(
            lambda obj: {'items': chain(
                [self.inject(
                    Link,
                    {'rel': str(obj['pk']), 'qs': self.get_location_qs()},
                    obj.get(first(first(self.props.guide))))],
                map(
                    lambda k: obj.get(k),
                    list(map(first, self.props.guide))[1:]))},
            self.props.objects))

    def render_table(self, component):
        return self.inject(
            component, {
                'thead': map(lambda c: c[1:], self.props.guide),
                'tbody': self.render_objects()})

    def render_queryset(self):
        table = self.render_table(SortableTable)
        if self.context['is_paginated']:
            return ''.join([table, self.inject(Paginator)])
        return table

    def render_actions(self):
        result = ''
        for action in self.props.actions:
            action['link']['qs'] = self.get_location_qs()
            result += self.inject(Action, action)
        return result

    def render_search(self):
        return '' if not self.props.search else self.inject(SearchLine)

    def render_filters(self):
        result = []
        for f in self.props.filters:
            if f['choices']:
                result.append(self.inject(QuerySetFilter, f))
        return result

    def render_panel(self):
        return ''.join(chain(*self.build(('actions', 'search', 'filters'))))

    def DOM(self):
        return ''.join(self.props.reference([
            self.inject(Column, {'md': 9, 'class': 'list-qs'}, self.render_queryset()),
            self.inject(Column, {'md': 3, 'class': 'list-panel'}, self.render_panel())]))



import string

from django import forms
from django.conf import settings
from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import get_hasher
from django.utils.text import capfirst

from era import _
from era.views import RedirectView, FormView, CollectionView, ObjectView
from era.utils.functools import pick

from .components import ResetNotification
from .decorators import login_required, role_required
from .mixins import AnonymousMixin, UserMixin, PasswordMixin, LoginMixin, WelcomeMixin, \
    InvitationMixin, SignMixin
from .models import Confirm


class LoginView(AnonymousMixin, UserMixin, LoginMixin, WelcomeMixin, FormView):
    validate_unique = False

    def get_actions(self):
        return [{
            'icon': 'sign-in',
            'title': _('login'),
            'level': 'success'
            }, {
            'icon': 'user-plus',
            'title': _('join'),
            'level': 'link',
            'link': 'registration'
            }, {
            'icon': 'unlock',
            'title': _('unlock'),
            'level': 'link',
            'link': 'reset'}]

    def prepare_form(self, form):
        form = super().prepare_form(form)
        form.fields['password'].widget = forms.PasswordInput()
        return form

    def process_valid(self, form, **kw):
        user = auth.authenticate(**form.cleaned_data)
        if user:
            if Confirm.objects.filter(user=user, key='registration').count():
                self.send_message('error', _('sorry, unconfirmed data'))
            else:
                return self.process_login(user)
        else:
            self.send_message('error', _('sorry, invalid credentials'))
        return self.reload()

    def get_success_redirect(self, **kw):
        return self.request.GET.get('next', super().get_success_redirect(**kw))


class RegistrationView(AnonymousMixin, SignMixin, FormView):
    extra_fields = ['name', 'password']
    validate_unique = False
    confirm = 'registration'
    success_redirect = 'confirm'
    success_message = _('confirmation data has been sent')

    def process_valid(self, form, **kw):
        user = self.get_user(form)
        if user:
            confirm = self.get_confirmation(user).first()
            if confirm:
                user.delete()
                confirm.delete()
            else:
                self.send_message('error', _('sorry, user with such data exists'))
                return self.reload()
        return super().process_valid(form=form, **kw)

    def save_form(self, form):
        form.instance.role = settings.USER_ROLES[-1].string
        super().save_form(form)


class ResetView(AnonymousMixin, SignMixin, FormView):
    validate_unique = False
    notification = ResetNotification
    notification_message = _('access restoration')
    success_redirect = 'unlock'

    def get(self, *args, **kw):
        self.send_message('info', _('please set new password and confirm it by code'))
        return super().get(*args, **kw)

    def prepare_form(self, form):
        form = super().prepare_form(form)
        form.fields['password'].label = capfirst(self.password_messages['new'])
        return form

    def process_valid(self, form, **kw):
        user = self.get_user(form)
        if user:
            self.get_confirmation(user).delete()
            return super().process_valid(form=form, **kw)
        else:
            self.send_message('error', _('sorry, invalid credentials'))
            return self.reload()

    def save_form(self, form):
        self.send_notification(form)


class ConfirmView(AnonymousMixin, LoginMixin, WelcomeMixin, FormView):
    model = Confirm
    fields = ('code', 'sign')
    repeat_url = 'registration'

    def check(self, code, sign):
        confirm = Confirm.objects.filter(code=get_hasher().encode(code, sign)).first()
        if confirm:
            confirm.user.password = confirm.sign
            confirm.user.save()
            user = confirm.user
            confirm.delete()
            return self.process_login(auth.authenticate(**dict(
                user.username_dict,
                password=sign)))
        else:
            self.send_message('error', _('sorry, invalid credentials'))
        return False

    def get(self, *args, **kw):
        data = pick(self.request.GET, 'code', 'sign')
        if 'code' in data and 'sign' in data:
            return self.check(**data) or self.navigate('confirm')
        return super().get(*args, **kw)

    def get_actions(self):
        result = [{'icon': 'check', 'title': _('submit'), 'level': 'success'}]
        if not 'code' in self.request.GET:
            result.append({
                'icon': 'refresh',
                'title': _('repeat'),
                'level': 'link',
                'link': self.repeat_url})
        return result

    def prepare_form(self, form):
        form = super().prepare_form(form)
        if 'code' in self.request.GET:
            form.fields['code'].widget = forms.HiddenInput(
                attrs={'value': self.request.GET['code']})
        else:
            form.fields['code'].label = capfirst(_('enter received code'))

        if 'sign' in self.request.GET:
            form.fields['sign'].label = capfirst(_('enter received password'))
        else:
            form.fields['sign'].label = capfirst(_('enter your password again'))
        form.fields['sign'].widget = forms.PasswordInput()
        return form

    def process_valid(self, form, **kw):
        return self.check(**pick(form.cleaned_data, 'code', 'sign')) or self.reload()


class UnlockView(ConfirmView):
    repeat_url = 'reset'


class UserView(InvitationMixin, ObjectView):
    decorators = [login_required, role_required(allow=['developer'])]
    extra_fields = ['role', 'name', 'password', 'access']
    form_props = {'class': 'condensed'}


class UsersView(CollectionView):
    decorators = [login_required, role_required(allow=['developer'])]
    model = get_user_model()
    list_display = (model.USERNAME_FIELD, 'role', 'name', 'access')
    list_filter = ('role', 'access')
    list_counters = ('role', )
    list_search = ('name', )
    default_state = {'filters': {'access': True}}


class ProfileView(UserMixin, PasswordMixin, LoginMixin, FormView):
    validate_unique = False
    decorators = [login_required]
    extra_fields = ['name', 'password']
    form_props = {'class': 'condensed'}

    def get_instance(self):
        return self.request.user

    def prepare_form(self, form):
        form = super().prepare_form(form)
        form.fields[self.model.USERNAME_FIELD].widget = forms.TextInput(
            attrs={'readonly': 'readonly'})
        return form

    def process_valid(self, form, **kw):
        if 'password' in form.changed_data:
            self.save_form(form)
            return self.process_login(auth.authenticate(**dict(
                self.request.user.username_dict,
                password=form.cleaned_data['password'])))
        return super().process_valid(**dict(kw, form=form))

    def get_success_message(self, **kw):
        return _('your profile was updated successfully')


class LogoutView(RedirectView):
    decorators = [login_required]
    permanent = True

    def get_redirect_url(self):
        auth.logout(self.request)
        self.pattern_name = 'index'
        return super().get_redirect_url()



from itertools import chain
from django.core.urlresolvers import reverse
from django.utils.text import capfirst
from urllib.parse import urlencode

from ..utils.functools import just, call, factual, reduce_dict, omit, pick, truthful
from ..utils.translation import _
from .library import register, Import, Component, ComplexComponent, Tag


@register.era
class Icon(Tag):
    el = 'i'
    inline = True

    def get_defaults(self):
        return {'fixed': True, 'nodelist': ''}

    def resolve_attrs(self):
        return {
            'class': ' '.join(chain(
                ['fa', 'fa-' + self.props.name],
                map(
                    lambda z: 'fa-' + z[0].format(self.props.pop(z[1])),
                    filter(
                        lambda z: z[1] in self.props,
                        zip(
                            ('{0}x', 'li', 'lg', 'fw', 'spin', 'rotate-{0}'),
                            ('size', 'list', 'large', 'fixed', 'spin', 'rotate'))))))}


class MarkedList(Tag):
    el = 'ul'

    def get_items(self):
        return self.props.items

    def get_nodelist(self):
        return ''.join(map(
            lambda item: self.inject(Tag, {'el': 'li'}, item),
            self.get_items()))


class IconicList(Component):
    def get_defaults(self):
        return {'icon': 'asterisk', 'heading': 'h4'}

    def DOM(self):
        return ''.join([
            '' if not self.props.get('title') else self.inject(
                Tag,
                {'el': self.props.heading},
                capfirst(self.props.title)),
            self.inject(
                Tag,
                {'el': 'ul', 'class': 'fa-ul'},
                ''.join(map(
                    lambda content: self.inject(
                        Tag,
                        {'el': 'li'},
                        ''.join([content, self.inject(
                            Icon, {'name': self.props.icon, 'list': True})])),
                    map(str, self.get_items()))))])


@register.era
class Break(Component):
    def get_defaults(self):
        return {'x': 1, 'ruler': False}

    def resolve_props(self):
        return {
            'x': int(self.props.x),
            'tag': '<{0}r />'.format(self.props.ruler and 'h' or 'b')}

    def DOM(self):
        if 'join' in self.props:
            return self.props.tag.join(self.props.join)
        return self.props.x * self.props.tag


@register.era
class ProgressBar(Tag):
    def get_defaults(self):
        return {'level': 'success'}

    def resolve_props(self):
        return {
            'class': 'progress',
            'value': str(int(self.props.value))}

    def DOM(self):
        return self.inject(
            Tag,
            {'attrs': {
                'class': 'progress-bar-' + self.props.level,
                'style': 'width: {0}%;'.format(self.props.value)}},
            self.props.value + '%')


@register.era
class Row(Tag):
    pass


@register.era
class Container(Tag):
    named = False

    def get_defaults(self):
        return {'fluid': False}

    def resolve_attrs(self):
        return {'class': 'container' + (self.props.fluid and '-fluid' or '')}


@register.era
class Column(Tag):
    def get_defaults(self):
        return {'constrict': False}

    def resolve_attrs(self):
        return {'class': ' '.join(reduce_dict(
            lambda k, v: 'col-{0}-{1}'.format(k, v),
            dict({'md': 12} if not self.props.constrict else {
                'md': 12 - self.props.constrict * 2,
                'md-offset': self.props.constrict
            }, **omit(self.props, 'nodelist', 'class', 'constrict'))))}


@register.era
class Link(Tag):
    el = 'a'
    named = False

    def get_defaults(self):
        return {
            'nodelist': '',
            'newtab': False,
            'reverse': True,
            'host': False,
            'url': 'index',
            'args': [],
            'kwargs': {},
            'rel': None,
            'qs': None}

    def resolve_props(self):
        if self.props.rel:
            return {
                'reverse': False,
                'url': '/'.join([
                    self.request.path.rstrip('/'),
                    self.props.rel])}
        return {}

    def get_url(self):
        return ''.join(factual([
            self.props.host and self.get_host(),
            (self.props.reverse and reverse or just)(
                self.props.url,
                **pick(self.props, 'args', 'kwargs')),
            self.props.qs and '?' + urlencode(self.props.qs)]))

    def get_nodelist(self):
        return self.props['nodelist'] or self.get_url()

    def resolve_attrs(self):
        return {
            'target': self.props.newtab and '_blank' or '_self',
            'href': self.get_url()}


@register.era
class Button(Tag):
    el = 'button'
    named = False

    def get_defaults(self):
        return {'level': 'primary', 'type': 'button', 'name': 'button'}

    def resolve_attrs(self):
        return dict(
            {'class': 'btn btn-' + self.props.level},
            **pick(self.props, 'type', 'onclick'))


@register.era
class Label(Tag):
    el = 'span'

    def get_defaults(self):
        return {'level': 'default'}

    def resolve_attrs(self):
        return {'class': 'label-' + self.props.level}


@register.era
class Alert(Tag):
    def get_defaults(self):
        return {'level': 'primary', 'dismiss': True}

    def resolve_attrs(self):
        return {'class': 'alert-' + self.props.level}

    def get_nodelist(self):
        return self.show(Tag, {'el': 'span'}, super().get_nodelist())

    def tweak(self):
        super().tweak()
        if self.props.dismiss:
            self.dom.prepend(
                self.inject(
                    Tag,
                    {'el': 'button', 'class': 'close', 'attrs': 
                        {'type': 'button', 'data-dismiss': 'alert'}},
                    self.inject(Icon, {'name': 'remove'})))


@register.era
class Well(Tag):
    def resolve_attrs(self):
        if 'size' in self.props:
            return {'class': 'well-' + self.props.size}
        return {}


@register.era
class Caption(Tag):
    el = 'span'

    def resolve_props(self):
        if isinstance(self.props.get('icon'), str):
            return {'icon': {'name': self.props.icon}}
        return {}

    def get_nodelist(self):
        if not 'icon' in self.props:
            return self.props.title
        return ''.join([
            self.inject(Icon, self.props.icon),
            self.inject(Tag, {'el': 'span'}, self.props.get(
                'title', self.props.nodelist))])


@register.era
class Panel(Tag):
    parts = ['title', 'body']

    def get_defaults(self):
        return {'level': 'default'}

    def resolve_attrs(self):
        return {'class': 'panel-' + self.props.level}

    def resolve_props(self):
        return {
            'class': 'panel',
            'has_heading': bool(self.props.get('title', self.props.get('caption')))}

    def render_heading(self, **kw):
        return self.inject(
            Tag, dict({'class': 'panel-heading'}, **kw), self.inject(
                Tag,
                {'el': 'h4', 'class': 'panel-title'},
                self.props.get('title') or self.inject(Caption, self.props.caption)))

    def render_body(self, **kw):
        return self.inject(
            Tag, dict({'class': 'panel-body'}, **kw), self.props.body)

    def get_nodelist(self):
        return ''.join(factual([
            self.props.has_heading and self.render_heading(),
            self.render_body()]))


class CollapsiblePanel(Panel):
    named = False

    def get_defaults(self):
        return dict(super().get_defaults(), collapse=None)

    def resolve_props(self):
        return dict(
            super().resolve_props(),
            href=''.join([self.props.prefix, str(self.props.pk)]))

    def render_heading(self, **kw):
        return super().render_heading(**{'attrs': {
            'class': 'collapsed' if not self.props.collapse else '',
            'href': '#' + self.props.href,
            'data-toggle': 'collapse',
            'data-parent': '#' + str(self.props.parent)}})

    def render_body(self, **kw):
        return self.show(
            Tag,
            {'attrs': {
                'id': self.props.href,
                'class': ' '.join(factual(
                    ['panel-collapse', 'collapse', self.props.collapse]))}},
            super().render_body())


class Accordion(Tag):
    panel = CollapsiblePanel

    def open_gen(self):
        yield True
        while True: yield False

    def get_defaults(self):
        return {
            'id': 'accordion',
            'class': 'panel-group',
            'open': self.open_gen()}

    def check_is_open(self, obj):
        return next(self.props.open)

    def get_nodelist(self):
        return ''.join(map(
            lambda obj: self.show(
                self.panel, dict({
                    'collapse': self.check_is_open(obj) and 'in',
                    'parent': self.props.id,
                    'prefix': self.props.prefix},
                    **pick(obj, 'pk', 'caption', 'body'))),
            self.get_objects()))


@register.era
class Navbar(Tag):
    el = 'nav'
    parts = ['head', 'brand', 'collapse', 'text']

    def get_defaults(self):
        return {
            'key': 'default',
            'container': False,
            'fixed': False,
            'margin': 20}

    def resolve_attrs(self):
        result = {'class': 'navbar-default'}
        if self.props.fixed:
            result['class'] = ' '.join([
                result['class'], 'navbar-fixed-{0}'.format(self.props.fixed)])
            if 'height' in self.props:
                result['style'] = 'min-height: {0}px;'.format(self.props.height)
        return result

    def get_nodelist(self):
        nodelist = ''.join(factual([
            self.inject(
                Tag,
                {'class': 'navbar-header'},
                ''.join([
                    self.props.head or self.inject(
                        Tag,
                        {'el': 'button', 'attrs': {
                            'class': 'navbar-toggle collapsed',
                            'data-toggle': 'collapse',
                            'data-target': '#' + self.props.key}},
                        self.inject(Icon, {'name': 'align-justify'})),
                    '' if not self.props.brand else self.inject(
                        Link,
                        {'url': '/', 'reverse': False, 'class': 'navbar-brand'},
                        self.props.brand)])),
            self.props.text and self.inject(
                Tag,
                {'el': 'p', 'class': 'navbar-text'},
                self.props.text),
            self.props.collapse and self.inject(
                Tag,
                {'attrs': {'id': self.props.key, 'class': 'navbar-collapse collapse'}},
                self.props.collapse)]))
        return nodelist if not self.props.container else self.inject(
            Container, {'fluid': self.props.container == 'fluid'}, nodelist)

    def tweak(self):
        if self.props.fixed:
            self.dom.before(self.inject(
                Tag, {'el': 'style'}, 'body {{padding-{0}: {1}px;}}'.format(
                    self.props.fixed, sum(map(int, [
                        self.props.get('height', 50),
                        self.props.margin])))))


class Table(Tag):
    el = 'table'

    def slice(self, seq):
        return list(seq)[slice(*call(
            getattr(self, 'get_slice', lambda: (None, None))))]

    def get_thead_items(self):
        return self.props.get('thead', [])

    def get_tbody_items(self):
        return self.props.get('tbody', [])

    def get_defaults(self):
        return {
            'striped': False,
            'bordered': False,
            'hover': True,
            'condensed': True,
            'responsive': True}

    def render_content(self, content, cell):
        if isinstance(content, bool):
            return self.inject(
                Icon,
                {'name': (content and 'check' or 'minus') + '-circle'})
        return content

    def render_items(self, items, cell='td'):
        return ''.join(map(
            lambda row: self.inject(
                Tag,
                truthful({'el': 'tr', 'class': row.get('level')}),
                ''.join(map(
                    lambda c: self.inject(
                        Tag,
                        {'el': cell},
                        self.render_content(c, cell)),
                    self.slice(row['items'])))),
            items))

    def resolve_attrs(self):
        return {
            'class': self.get_class_set(
                'striped',
                'bordered',
                'hover',
                'condensed',
                'responsive',
                prefix='table')}

    def get_nodelist(self):
        body = list(self.get_tbody_items())
        if not len(list(body)):
            return _('(None)')
        return ''.join([
            self.inject(
                Tag,
                {'el': 'thead'},
                self.render_items([{'items': self.get_thead_items()}], cell='th')),
            self.inject(
                Tag,
                {'el': 'tbody'},
                self.render_items(body))])



from functools import reduce, wraps
from itertools import chain, tee
from random import choice
from string import digits
from django.utils.functional import curry


def just(*x, **kw):
    return len(x) and x[0] or kw

def call(fn):
    return fn()

def swap(a, fn):
    return fn(a)

def unpack_args(fn):
    return lambda t: fn(*t)

def throw(exception):
    raise exception

def unidec(fnx):
    '''
    @unidec
    def render(view, request, flag=True):
        pass

    @render
    def first_view(request):
        pass

    @render(flag=False)
    def second_view(request):
        pass
    '''

    return lambda *ax, **kx: (
        wraps(ax[0])(lambda *ay, **ky: fnx(ax[0], *ay, **ky)) \
        if len(ax) == 1 and not kx and callable(ax[0]) else \
        lambda fny: wraps(fny)(lambda *ay, **ky: \
        fnx(fny, *ay, **dict(kx, **ky)) if not ax else throw(
            TypeError('wrapper get *args'))))


def pluck(l, k):
    return list(map(
        lambda o: o.get(k) if isinstance(o, dict) \
        else reduce(lambda x, y: getattr(x, y, None), k.split('.'), o), l))

def first(seq):
    return list(seq)[0]

def select(i, l):
    return l[i - len(l) * int(i / len(l))]

def factual(l):
    return list(filter(bool, l))

def separate(fn, lx):
    return map(list, map(
        unpack_args(
            lambda i, ly: filter(
                lambda el: bool(i) == fn(el),
                ly)),
        enumerate(tee(lx, 2))))


def case(k, *dl, **kw):
    return dict(len(dl) and dl[0] or {}, **kw).get(k)

@unidec
def dict_copy(fn, d, *a):
    return {k: v for k, v in d.items() if fn(k, v, *a)}

@dict_copy
def pick(k, v, *a):
    return k in a

@dict_copy
def omit(k, v, *a):
    return not k in a

@dict_copy
def truthful(k, v, *a):
    return bool(v)

def filter_dict(fn, d):
    return dict_copy(fn)(d)

def reduce_dict(fn, d):
    return map(unpack_args(fn), d.items())

@unidec
def dict_map(fnx, fny, d):
    return dict(reduce_dict(lambda k, v: fnx(k, v, fny), d))

@dict_map
def map_keys(k, v, fn):
    return (fn(k), v)

@dict_map
def map_values(k, v, fn):
    return (k, fn(v))


def avg(*args):
    return reduce(lambda x, y: x + y, args) / len(args)

def random_str(length=6, source=digits):
    return ''.join(map(lambda i: choice(source), range(0, length)))



from era import package_patterns, admin_urls, view_url
from django.contrib.auth import get_user_model


urlpatterns = package_patterns(
    __package__,
    'login',
    'logout',
    'profile',
    'registration',
    'reset',
    'confirm',
    'unlock')

urlpatterns += package_patterns(
    __package__, *admin_urls(get_user_model()))



from itertools import chain
from django.forms import widgets
from django.forms.widgets import CheckboxInput
from django.template.defaulttags import CsrfTokenNode

from ..forms import EmptyWidget
from ..utils.functools import call, unpack_args, factual, pluck, separate, \
    case, pick, omit, truthful
from .library import register, Component, Tag, ScriptedTag
from .markup import Row, Column, Table, Link, Button, Caption, Panel, Icon


class WidgetCaseMixin:
    def get_widgets(self, *args):
        return list(map(lambda x: getattr(widgets, x), args))

    def check_widget_in(self, *args):
        return self.props.field.field.widget.__class__ in self.get_widgets(*args)

    @property
    def is_text_input(self):
        return getattr(
            self.props.field.field.widget,
            'is_text_input',
            self.check_widget_in(
                'TextInput', 'NumberInput', 'EmailInput',
                'URLInput', 'PasswordInput', 'Textarea', 'Select'))

    @property
    def have_placeholder(self):
        return self.check_widget_in('TextInput', 'Textarea', 'NumberInput')


class RequiredAttrMixin:
    @property
    def need_required(self):
        return self.props.set_required and self.props.field.field.required


class Input(WidgetCaseMixin, RequiredAttrMixin, Component):
    def DOM(self):
        return self.props.field.as_widget()

    def tweak(self):
        super().tweak()
        if self.is_text_input:
            self.dom.add_class('form-control')
        if self.need_required:
            self.dom.add_attr('required')
        if self.props.inline:
            self.dom.add_attr('placeholder', self.props.field.label)
        elif self.have_placeholder:
            self.dom.add_attr('placeholder', self.props.field.help_text)


class Label(RequiredAttrMixin, Tag):
    el = 'label'
    named = False

    def resolve_attrs(self):
        return {
            'for': self.props.field.id_for_label,
            'class': ' '.join([
                'control-label',
                self.need_required and 'required' or ''])}

    def get_nodelist(self):
        return ''.join(map(str, [
            self.props.get('nodelist', ''),
            self.props.field.label]))


class Group(Tag):
    named = False

    @property
    def is_valid(self):
        return self.props.valid

    def get_defaults(self):
        return {'valid': None, 'inline': False}

    def resolve_attrs(self):
        return {'class': ' '.join(factual([
            'form-group',
            case(self.is_valid, {
                True: 'has-success',
                False: 'has-error', None: ''})]))}


class Field(WidgetCaseMixin, Group):
    def get_defaults(self):
        return dict(super().get_defaults(), **{
            'set_label': True,
            'set_required': True})

    @property
    def is_valid(self):
        return (False if self.props.field.errors else True) \
            if hasattr(self.props.field.form, 'cleaned_data') else None

    def inject_control(self, control, nodelist=None):
        return self.inject(
            control, pick(self.props, 'field', 'inline', 'set_required'), nodelist)

    def get_help_block(self):
        if self.props.field.help_text and not self.have_placeholder:
            return self.inject(
                Tag,
                {'el': 'span', 'class': 'help-block'},
                self.props.field.help_text)

    def get_nodelist(self):
        if isinstance(self.props.field.field.widget, CheckboxInput):
            if not self.props.set_label:
                return self.inject_control(Input)
            return self.inject(
                Tag,
                {'class': 'checkbox'},
                self.inject_control(Label, self.inject_control(Input)))
        return ''.join([
            '' if (self.props.inline or not self.props.set_label) \
                else self.inject_control(Label),
            self.inject_control(Input),
            '' if self.is_valid is not False else self.inject(
                Tag,
                {'el': 'span', 'class': 'help-block'},
                '<br />'.join(map(str, self.props.field.errors))),
            self.get_help_block() or ''])


class FieldsetMixin:
    def render_fieldset(self, form, **props):
        [hidden, fields] = separate(lambda f: not f.is_hidden, form)
        return map(unpack_args(
            lambda i, field: ''.join([
                '' if i else ''.join(map(lambda f: f.as_widget(), hidden)),
                self.inject(Field, dict({'field': field}, **props))])),
            enumerate(fields))


class Formset(Table, FieldsetMixin):
    def get_slice(self):
        have_instances = any([f.instance.pk for f in self.props.formset.forms])
        return (0, -1 if self.props.formset.can_delete and not have_instances else None)

    def get_thead_items(self):
        return factual(map(
            lambda f: not f.is_hidden and f.label,
            self.props.formset.forms[0]))

    def get_form_fields(self, form):
        if self.props.formset.can_delete and not form.instance.pk:
            form.fields['DELETE'].widget = EmptyWidget()
        return tuple(form)

    def get_tbody_items(self):
        return map(
            lambda form: {'items': self.render_fieldset(
                self.get_form_fields(form),
                inline=False,
                set_label=False,
                set_required=False)},
            self.props.formset.forms)


class Action(Tag):
    def get_button_props(self):
        if not pick(self.props, 'link', 'onclick'):
            return dict({'type': 'submit'}, **pick(self.props, 'level'))
        return pick(self.props, 'level', 'onclick')

    def get_link_props(self):
        if isinstance(self.props.link, str):
            return {'url': self.props.link}
        return self.props.link

    def get_nodelist(self):
        result = self.inject(
            Button, self.get_button_props(), self.inject(
                Caption, pick(self.props, 'icon', 'title')))
        if 'link' in self.props:
            return self.inject(Link, self.get_link_props(), result)
        return result


@register.era
class Form(ScriptedTag, FieldsetMixin):
    el = 'form'
    inline = True
    named = False

    def get_defaults(self):
        return {
            'prefix': '',
            'action': self.request.get_full_path(),
            'method': 'POST',
            'novalidate': False,
            'inline': False,
            'panels': False,
            'spinner': False,
            'title': '',
            'splitters': [],
            'relations': [],
            'formsets': []}

    def get_context_prop(self, prop, default=None):
        return self.context.get(
            '_'.join(factual([self.props.prefix, prop])),
            default)

    def resolve_enctype(self):
        if any(map(
            lambda x: x.is_multipart(),
            chain(
                [self.props.form],
                pluck(self.props.relations, 'form'),
                self.props.formsets))):
            return 'multipart/form-data'
        return 'application/x-www-form-urlencoded'

    def resolve_props(self):
        return dict(
            truthful(dict(map(
                lambda p: (p, self.get_context_prop(p)),
                ['form', 'relations', 'formsets']))),
            **self.get_context_prop('props', {}))

    def split_fields(self):
        result = [[]]
        for field in self.props.form:
            result[-1].append(field)
            if field.name in self.props.splitters:
                result.append([])
        return result

    def render_panel(self, **kw):
        if self.props.panels:
            if kw['body']:
                return self.inject(Panel, kw)
        return kw['body']

    def render_fields(self):
        if self.props.splitters:
            columns = self.split_fields()
            result = ''.join(map(
                lambda column: self.inject(
                    Column,
                    {'md': int(12 / len(columns)), 'xs': 12},
                    ''.join(self.render_fieldset(
                        column,
                        **pick(self.props, 'inline')))),
                columns))
        else:
            result = ''.join(self.render_fieldset(
                self.props.form,
                **pick(self.props, 'inline')))
        return self.render_panel(caption={'title': self.props.title}, body=result)

    def render_relations(self):
        return ''.join(map(
            lambda relation: self.render_panel(
                caption={'title': relation['field'].verbose_name},
                body=''.join(
                    self.render_fieldset(
                        relation['form'],
                        set_required=relation['required']))),
            self.props.relations))

    def render_formsets(self):
        return ''.join(map(
            lambda formset: self.render_panel(
                caption={'title': formset.model._meta.verbose_name_plural},
                body=''.join([
                    str(formset.management_form),
                    self.inject(Formset, {'formset': formset})])),
            self.props.formsets))

    def render_actions(self):
        return self.inject(
            Group,
            {'class': 'actions'},
            ''.join(map(
                lambda action: self.inject(Action, action),
                self.props.actions)))

    def resolve_script(self):
        result = pick(self.props, 'spinner', 'action')
        if self.props.spinner:
            result['spinner'] = self.inject(Icon, {
                'name': self.props.spinner,
                'spin': True,
                'large': True})
        return result

    def get_nodelist(self):
        return ''.join(chain(
            [CsrfTokenNode().render(self.context)],
            map(
                lambda content: '' if not content else \
                    self.props.inline and content or \
                    self.inject(Row, {}, content),
                self.build(('fields', 'relations', 'formsets', 'actions')))))

    def resolve_attrs(self):
        return dict(
            self.props.novalidate and {'novalidate': None} or {}, **{
            'class': 'form',
            'method': self.props.method,
            'action': self.props.action,
            'enctype': self.resolve_enctype()})
